#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <glib.h>

#include <bmpx/metadata.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>

typedef struct _LCMFS LCMFS;

struct _LCMFS {
    int		   i_size;
    GHashTable	  *i;

    gchar	  *index;
    gchar	  *records;

    FILE	  *f_index; 
    FILE	  *f_records; 
};

static LCMFS *lcmfs;

void
lcmfs_init ()
{
  lcmfs = g_new0 (LCMFS,1);
  
  lcmfs->i_size = 
      0;
  lcmfs->i = 
      g_hash_table_new (g_str_hash, g_str_equal);
}

void
lcmfs_open (const gchar *lcmfs_base)
{
  lcmfs->index =
    g_strconcat (lcmfs_base, ".lif", NULL); // LCMFS Index  File
  lcmfs->records = 
    g_strconcat (lcmfs_base, ".lrd", NULL); // LCMFS Record Data

  // Open Index
  lcmfs->f_index = fopen (lcmfs->index, "r");
  if (lcmfs->f_index)
    {
      // Read index
      gint n;

      fread (&lcmfs->i_size, sizeof(int), 1, lcmfs->f_index);
      for (n = 0; n < lcmfs->i_size; n++)
	{
	  uint16_t key_size;
	  gint     offset;
	  gchar   *pkey;

	  fread (&key_size, sizeof(uint16_t), 1, lcmfs->f_index);
	  pkey = g_malloc0 (key_size+1);
	  fread (pkey, sizeof(char), key_size, lcmfs->f_index);
	  fread (&offset, sizeof(gint), 1, lcmfs->f_index);

	  g_hash_table_insert (lcmfs->i, pkey, GINT_TO_POINTER (offset)); 
	}

      fclose (lcmfs->f_index);
    } 

  // Open Records
  lcmfs->f_records = fopen (lcmfs->records, "a+");
  // FIXME: Check errno after lockf ()
}

gboolean
lcmfs_record_exists (const gchar *pkey)
{
  return g_hash_table_lookup (lcmfs->i, pkey) ? TRUE : FALSE;
}

void
record_read_mem (HTuple *record, unsigned char **data)
{
  gint n;

  for (n = 0; n < BMP_DATUM_NONE ; n++)
    {
      gint *metadatum_id_ptr;
      gint metadatum_id;
      
      metadatum_id_ptr = (gint *)(*data);
      metadatum_id = *metadatum_id_ptr;
      *data += sizeof(gint);
 
      if (bmp_metadata_get_type (metadatum_id) == G_TYPE_INT)
        {
          gint *value;
 
          value = (gint*)(*data);
 
	  h_tuple_insert (record,
			  bmp_metadata_get_id_static (metadatum_id),
			  G_TYPE_INT,
			  GINT_TO_POINTER(*value));
 
          *data += sizeof(gint);
        }
      else if (bmp_metadata_get_type (metadatum_id) == G_TYPE_STRING)
        {
          char *value;
 
          value = (char *)(*data);
 
	  h_tuple_insert (record,
			  bmp_metadata_get_id_static (metadatum_id),
			  G_TYPE_STRING,
			  value); 
 
           /* string length + terminating 0 */
          *data += strlen(value) + 1;
        }
    }
}

void
lcmfs_read (BmpMetadataCache *cache)
{
  unsigned char *current_ptr, *current_ptr_save;
  FILE		*fp;
  gint		 n, ret, fd;

  struct stat	 st;

  ret = stat (lcmfs->records, &st);
  if (ret == -1)
    {
      /* File does not exits? */
      g_message ("LCMFS Records do not exist.");
      return;
    }

  fp = fopen (lcmfs->records, "r");

  if (!fp)
    {
      g_message ("Could not open LCMFS Records.");
      return;
    }

  fd = fileno (fp);

  current_ptr = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
  current_ptr_save = current_ptr;
  fclose(fp);

  if (!current_ptr)
    {
      g_message ("Could not mmap LCMFS Records.");
      return;
    }

  for (n = 0; n < lcmfs->i_size; n++)
    {
      HTuple *record;

      record = bmp_metadata_tuple_create_default ();
      record_read_mem (record, &current_ptr);

      _bmp_metadata_cache_tuple_inserted (cache, record); 
    }

  munmap(current_ptr_save, st.st_size);
}

gboolean
lcmfs_record_add (HTuple *record)
{
  HTupleElem *elem_pkey, *elem;

  gint	    new_offset, n = 0;

  // Just for the case we check if the record exists
  elem_pkey = h_tuple_get (record, (gpointer)bmp_metadata_get_id_static (BMP_METADATA_PKEY));
  if (lcmfs_record_exists (elem_pkey->value))
      return FALSE;

  // Seek to the end of the records file
  fseek (lcmfs->f_records, 0L, SEEK_END);
  new_offset = ftell (lcmfs->f_records);

  for (n = 0; n < BMP_DATUM_NONE; n++)
    {
      elem = h_tuple_get (record, bmp_metadata_get_id_static (n)); 

      if (elem->type == G_TYPE_INT)
	{
	  gint v;

	  v = GPOINTER_TO_INT (elem->value);

	  fwrite (&n, sizeof(gint), 1, lcmfs->f_records);
          fwrite (&v, sizeof(gint), 1, lcmfs->f_records);
	}
      else if (elem->type == G_TYPE_STRING)
	{
	  fwrite (&n, sizeof(gint), 1, lcmfs->f_records);
          fwrite (elem->value, sizeof(char), strlen(elem->value)+1, lcmfs->f_records);
	}
    }

  g_hash_table_insert (lcmfs->i, g_strdup (elem_pkey->value), GINT_TO_POINTER (new_offset));
  lcmfs->i_size++;

  return TRUE;
}

static void
_write_index_record (gpointer key, gpointer value, gpointer data)
{
  uint16_t  key_size;
  gchar    *pkey   = (gchar*)key;
  gint      offset = GPOINTER_TO_INT (value);

  key_size = strlen (pkey);

  fwrite (&key_size, sizeof(uint16_t), 1, lcmfs->f_index); 
  fwrite (pkey, sizeof(char), key_size, lcmfs->f_index);
  fwrite (&offset, sizeof(gint), 1, lcmfs->f_index);
}

void
lcmfs_close ()
{
  lcmfs->f_index = fopen (lcmfs->index, "w");
  if (lcmfs->f_index)
    {
      // Write index
      fwrite (&lcmfs->i_size, sizeof(int), 1, lcmfs->f_index);
      g_hash_table_foreach (lcmfs->i, (GHFunc)_write_index_record, NULL);
      fclose (lcmfs->f_index);
    } 
  else
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to write index file", G_STRLOC);
      // ERROR
    }

  fclose (lcmfs->f_records);
}
